#if !defined(LD_OS_CORE_H_)
#define LD_OS_CORE_H_

typedef struct OS_Handle OS_Handle;
struct OS_Handle {
	U64 v[1];
};

// Virtual memory

function U64 VM_PageSize();
function U64 VM_AllocationGranularity();

function void *VM_Reserve(U64 size);
function B32   VM_Commit(void *base, U64 size);
function void  VM_Decommit(void *base, U64 size);
function void  VM_Release(void *base, U64 size);

#include "filesystem.h"

#endif  // LD_OS_CORE_H_
